DEFAULT_BUCKET_WRITE: str = 'sagemaker-fmbench-write'
DEFAULT_BUCKET_READ: str = 'sagemaker-fmbench-read'
DEFAULT_LOCAL_WRITE: str = 'fmbench-write'
DEFAULT_LOCAL_READ: str = 'fmbench-read'
MAX_REQ_CNT: int = 105
MIN_REQ_CNT: int = 5
